import decimal

from django.contrib.auth.models import User
from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.translation import ugettext as _
from treenode.models import TreeNodeModel
from django.utils.timezone import now


class TrackCode(models.Model):
    node = models.ForeignKey('MarketerNode', null=True, blank=True, on_delete=models.CASCADE)
    code = models.CharField(max_length=64, db_index=True, unique=True)
    created_at = models.DateTimeField(auto_now_add=True)
    valid_from = models.DateField(null=True, blank=True)
    valid_until = models.DateField(null=True, blank=True)
    USAGE_CHOICES = ( ("always", _("always",)),
                      ("useronce", _("Nx / user")),
                      ("once", _("Nx / user")))
    usage = models.CharField(max_length=8, choices=USAGE_CHOICES, default="useronce")
    max_n = models.PositiveIntegerField(default=1, help_text=_("The maximum number of codes applied to usage choices"))
    discount = models.DecimalField(default=10, decimal_places=2, max_digits=5)
    DISCOUNT_TYPE_CHOICES = ( ("perc", "%"),
                              ("CZK", "CZK"))
    discount_type = models.CharField(max_length=4, choices=DISCOUNT_TYPE_CHOICES, default="perc", verbose_name="dt")

    def __str__(self):
        return self.code

    @property
    def num_used(self):
        return self.codeuse_set.count()


class MarketerNode(TreeNodeModel):
    treenode_display_field = 'name'
    name              = models.CharField(max_length=256)
    owner             = models.ForeignKey(User, on_delete=models.CASCADE, null=True)
    SPLIT_TYPE_CHOICES = ( ( "fixed", _("fixed"),),
                           ( "treesplit", _("tree split")) )
    split_type        = models.CharField(max_length=16, choices=SPLIT_TYPE_CHOICES, default="fixed", help_text=_("How to split, fixed amount (1 level), or tree split (recursively sent up the tree)"))
    split_cut         = models.DecimalField(default=100.0, decimal_places=2, max_digits=5, help_text=_("% of split between THIS one and other"))
    can_have_children = models.BooleanField(default=True)

    @property
    def node_cut(self):
        return self.split_cut

    @property
    def sub_cut(self):
        return 100-self.split_cut

    def max_available(self, value=1.0):
        value = decimal.Decimal(value)
        for n in self.ancestors:
            if n.split_type == "fixed":
                value*= n.sub_cut/100
        return value

    @property
    def max_discount(self):
        return self.max_available(100)

    def my_cut(self, value=1.0):
        if self.parent == None:
            return value * self.split_cut/100
        value = self.max_available(value)
        if self.split_type == "treesplit" and self.parent and self.parent.split_type== "treesplit":
            return value * self.sub_cut/100
        return value

    def provision_min(self, value=1.0):
        if self.split_type == "treesplit":
            mc = self.max_available(value)
            for i in range(0,self.depth):
                mc-= mc*self.sub_cut/100
            if self.parent and self.parent.split_type!= "treesplit":
                return mc
            else:
                return mc*self.sub_cut/100
        else:
            value = decimal.Decimal(value)
            for n in self.ancestors:
                value*=n.sub_cut/100
            if self.parent:
                return value*self.node_cut/100
            else: #top level
                return value*self.sub_cut/100

    def save(self, *args, **kwargs):
        if self.parent and self.parent.split_type == "treesplit":
            self.split_type = self.parent.split_type
            self.split_cut = self.parent.split_cut
        for n in self.children:
            n.split_type = self.split_type
            n.split_cut = self.split_cut
            n.save()
        return super(MarketerNode, self).save(*args, **kwargs)


    def get_tree_split_ancestors(self):
        ancs = []
        if self.split_type == "treesplit":
            # yield self
            for anc in self.ancestors[::-1]:
                if anc.split_type != "treesplit":
                    break
                ancs.append(anc)
        return ancs


    def subpath(self, subnode):
        if subnode == self:
            return []
        ret = [subnode]
        for x in subnode.ancestors[::-1]:
            if x==self:
                break
            ret.append(x)
        return ret[::-1]


    def split_the_spoils(self, amount, cash_node, discount=0.0):
        ''' Splits the amount based on the tree. self is cashing and casher_node is the one who made the deal'''
        amount = decimal.Decimal(amount)
        if not self.parent:
            return amount*(self.split_cut/100)
        cut_for_node = cash_node.my_cut(100) - discount
        cash_node_cut = amount*(cut_for_node/100)
        if self==cash_node:
            return cash_node_cut
        currnode = cash_node
        if self.split_type == "treesplit":
            tree_split_ancestors = cash_node.get_tree_split_ancestors()
            upcut = cash_node.max_available(100)

            for x in tree_split_ancestors:
                upcut-= upcut*self.sub_cut/100
                if x == self and self.parent.split_type == "treesplit":
                    return amount * upcut / 100 * x.sub_cut / 100
            return amount*upcut/100
        else:
            return self.provision_min(amount)


    def get_provisions_for_period(self, start, end):
        ''' Generates report of code usage in the timeframe '''
        my_provisions = []
        scope = [self] + self.descendants
        subids = [d.id for d in self.descendants]
        for cu in CodeUse.objects.filter(code__node__in=scope).filter(used_at__gt=start, used_at__lte=end):
            cash_node = cu.code.node
            for_me = self.split_the_spoils(cu.amount, cash_node, cu.code.discount)
            my_provisions.append( (cu, for_me, self.subpath(cash_node)) )
        return my_provisions

    def get_sum_provisions_for_period(self, start, end):
        ''' Gets the total provisions for this node, from this node and its subnodes'''
        total = decimal.Decimal(0.0)
        for cu, provision, subpath in self.get_provisions_for_period(start, end):
            total+= provision
        return total

    def get_provisions_tree_for_period(self, start, end):
        ''' return list of (node, depth, individual provision, individual inherited provision, node's inherited provision)'''
        scope = [self] + self.descendants
        accumulator = {}
        for node in scope:
            if node not in accumulator:
                accumulator[node] = decimal.Decimal(0.0)
        for cu, provision, subpath in self.get_provisions_for_period(start, end):
            accumulator[cu.code.node]+= provision
        ret = [ ( self, 0, accumulator[self] ) ]
        for x in self.descendants:
            ret.append( (x, x.level-self.level, accumulator[x]) )
        return ret

class CodeUse(models.Model):
    code = models.ForeignKey(TrackCode, on_delete=models.DO_NOTHING)
    content_type = models.ForeignKey(ContentType, on_delete=models.DO_NOTHING, null=True)
    object_id = models.PositiveIntegerField(null=True)
    content_object = GenericForeignKey('content_type', 'object_id')
    user = models.ForeignKey(User, null=True, on_delete=models.DO_NOTHING)
    used_at = models.DateTimeField(auto_now=True, db_index=True)
    amount = models.DecimalField(default=10, decimal_places=2, max_digits=10)

    def can_be_used(self):
        ''' this method should be called before save '''
        if self.code.valid_from and self.code.valid_from > now().date():
            return False
        if self.code.valid_until and self.code.valid_until <= now().date():
            return False
        if self.code.usage == "once":
            if CodeUse.objects.filter(code=self.code).count() >= self.code.max_n:
                return False
        elif self.code.usage == "useronce":
            if CodeUse.objects.filter(code=self.code, user=self.user).count() >= self.code.max_n:
                return False
        #TODO can be applied to target
        return True

    @classmethod
    def test_use(cls, code_str, user, target=None, amount=0):
        ''' Test if the string code can be used for the desired user
        :param code_str - string of the discount code
        :param user - the Django user to map to
        :param target - the object this discount is referring to (usually the Order)
        :param amount - the original amount to calculate the discount '''
        try:
            code = TrackCode.objects.get(code__iexact=code_str)
        except TrackCode.DoesNotExist:
            return False, decimal.Decimal(0.0)
        cu = CodeUse(code=code, content_object=target, user=user, amount=amount)
        return cu.can_be_used(), cu.discounted_amount

    @classmethod
    def use(cls, code_str, user, target=None, amount=0):
        ''' Test if the string code can be used for the desired user
        :param code_str - string of the discount code
        :param user - the Django user to map to
        :param target - the object this discount is referring to (usually the Order)
        :param amount - the original amount to calculate the discount '''
        code = TrackCode.objects.get(code__iexact=code_str)
        cu = CodeUse(code=code, content_object=target, user=user, amount=amount)
        assert cu.can_be_used(), "This code can not be used"
        cu.save()

    def get_provision(self):
        ''' Gets the provision for this node '''
        cash_node = self.code.node
        return cash_node.split_the_spoils(self.amount, cash_node, self.code.discount) #TODO discount type

    def get_ancestor_provisions(self):
        cash_node = self.code.node
        discount = self.code.discount
        for node in cash_node.ancestors:
            yield (node, node.split_the_spoils(self.amount, cash_node, discount))
        yield (cash_node, cash_node.split_the_spoils(self.amount, cash_node, self.code.discount))

    @property
    def discounted_amount(self):
        self.amount = decimal.Decimal(self.amount)
        return self.amount*decimal.Decimal(self.code.discount)/100

    @property
    def control_sum(self):
        ''' Sum all after spliting the spoils. (must match the original amount) '''
        sum = decimal.Decimal(0)
        for node, split in self.get_ancestor_provisions():
            sum+= split
        sum+= self.discounted_amount
        return sum


class DummyTarget(models.Model):
    name = models.CharField(max_length=2048)
    amount = models.DecimalField(decimal_places=2, max_digits=10)
    code_used = GenericRelation(CodeUse)

    def __str__(self):
        return "DummyTarget %s:%s" % (self.pk, self.name)


from django.contrib import admin
from django.db.models import Q
from django.utils.safestring import mark_safe
from treenode.admin import TreeNodeModelAdmin
from treenode.utils import split_pks
from mlmtools.forms import MarketerNodeForm
from mlmtools.models import MarketerNode, TrackCode
from django.utils.translation import ugettext as _



class TrackCodeInline(admin.TabularInline):
    model = TrackCode
    extra = 0
    #TODO prevent deletion or discount change after CodeUse exists


@admin.register(MarketerNode)
class MarketerNodeAdmin(TreeNodeModelAdmin):
    list_display = ('name','owner','level','format_split_type','format_split','format_provision','format_my_cut')
    form = MarketerNodeForm
    inlines = [TrackCodeInline]
    MAX_SUB = 4

    def format_my_cut(self, obj):
        return "%.2f%%" % obj.my_cut(100)
    format_my_cut.short_description = _("my cut")

    def format_provision(self, obj):
        return mark_safe("%.2f%% - %.2f%%" % (obj.provision_min(100), obj.max_available(100) ))
    format_provision.short_description = _("provision")

    def format_split(self, obj):
        f = "%.0f/%.0f" % (obj.node_cut, obj.sub_cut)
        if obj.split_type=="treesplit" and obj.parent.split_type=="treesplit":
            return mark_safe("<span style='color: #999;'>"+f+"</span>")
        else:
            return f
    format_split.short_description = _("split")

    def format_split_type(self, obj):
        if obj.split_type=="treesplit" and obj.parent.split_type=="treesplit":
            return mark_safe("<span style='color: #999;'>"+obj.split_type+"</span>")
        else:
            return obj.split_type
    format_split_type.short_description = _("split type")

    def _get_treenode_display_mode(self, request, obj):
        if request.user.is_superuser:
            self.treenode_display_mode = TreeNodeModelAdmin.TREENODE_DISPLAY_MODE_ACCORDION
        else:
            self.treenode_display_mode = TreeNodeModelAdmin.TREENODE_DISPLAY_MODE_INDENTATION
        return self.treenode_display_mode


    def get_queryset(self, request):
        qs = super(MarketerNodeAdmin, self).get_queryset(request)
        if request.user.is_superuser:
            return qs
        else:
            q = Q(owner_id=request.user.id)
            for i in range(1, self.MAX_SUB):
                q = q | Q( **{"tn_parent__"*i+"owner_id" : request.user.id} )
            return qs.filter(q)


    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if not request.user.is_superuser:
            if db_field.name == "tn_parent":
                exclude_pks = []
                if self.instance:
                    obj = self.instance
                    if obj.pk:
                        exclude_pks += [obj.pk]
                        exclude_pks += split_pks(obj.tn_descendants_pks)
                kwargs["queryset"] = self.get_queryset(request).exclude(pk__in=exclude_pks)
                return db_field.formfield(**kwargs)

        return super(MarketerNodeAdmin, self).formfield_for_foreignkey(db_field, request, **kwargs)


    def get_form(self, request, obj=None, **kwargs):
        self.instance = obj
        return super(MarketerNodeAdmin, self).get_form(request, obj, **kwargs)





# Generated by Django 3.1.6 on 2021-02-19 08:36

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('mlmtools', '0002_marketernode'),
    ]

    operations = [
        migrations.AddField(
            model_name='marketernode',
            name='owner',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='trackcode',
            name='owner',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
    ]



# Generated by Django 3.1.6 on 2021-02-19 14:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('mlmtools', '0006_auto_20210219_1421'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='marketernode',
            name='node_cut',
        ),
        migrations.RemoveField(
            model_name='marketernode',
            name='sub_cut',
        ),
        migrations.AddField(
            model_name='marketernode',
            name='split_cut',
            field=models.DecimalField(decimal_places=2, default=100.0, help_text='% of split between this one and sub nodes', max_digits=5),
        ),
    ]



import decimal
from datetime import datetime
from decimal import Decimal

from django.contrib.auth.models import User
from django.test import TestCase

from mlmtools.models import MarketerNode, TrackCode, CodeUse, DummyTarget


class ProvisionIntegrityTestCase(TestCase):
    DS,DU = datetime.fromisoformat("1990-01-01"), datetime.fromisoformat("2100-01-01")
    def setUp(self):
        self.user = User.objects.create_user(username='testuser', password="12345")
        self.top_node = MarketerNode.objects.create(name="Top Base", split_type="fixed", split_cut=60,
                                               owner=self.user )
        self.l2_marketeri1 = MarketerNode.objects.create(name="Marketeri 1", split_type="fixed", split_cut=25,
                                                    owner=self.user, tn_parent=self.top_node )
        self.l3_sub1 = MarketerNode.objects.create(name="second sub1", split_type="treesplit", split_cut=25,
                                              owner=self.user, tn_parent=self.l2_marketeri1)
        self.l4_sub = MarketerNode.objects.create(name="L4 sub", split_type="treesplit", split_cut=25,
                                            owner=self.user, tn_parent=self.l3_sub1)
        self.l5_sub = MarketerNode.objects.create(name="L5 sub", split_type="treesplit", split_cut=25,
                                            owner=self.user, tn_parent=self.l4_sub)
        self.l2_marketeri2 = MarketerNode.objects.create(name="Marketeri 2", split_type="fixed", split_cut=0,
                                                    owner=self.user, tn_parent=self.top_node)

        self.test_client = User.objects.create_user(username='testclient', password="12345")
        self.test_code_l5_sub = TrackCode.objects.create(code="L5SUB1", usage="useronce", discount_type="perc", discount=0,
                                                         node=self.l5_sub)
        self.dummy_order1 = DummyTarget.objects.create(name="dummy target1", amount=1000.0)

        self.test_code_l5_sub2 = TrackCode.objects.create(code="L5SUB2", usage="useronce", discount_type="perc", discount=10,
                                                          node=self.l5_sub)
        self.test_code_l4_sub = TrackCode.objects.create(code="L4SUB1", usage="useronce", discount_type="perc", discount=10,
                                                         node=self.l4_sub)
        self.cu_simple_treesplit_no_discount = CodeUse.objects.create(code=self.test_code_l5_sub, content_object=self.dummy_order1, user=self.test_client, amount=self.dummy_order1.amount)
        self.cu_simple_treesplit_discount = CodeUse.objects.create(code=self.test_code_l5_sub2, content_object=self.dummy_order1, user=self.test_client, amount=self.dummy_order1.amount)
        self.cu_simple_treesplit_l4_discount = CodeUse.objects.create(code=self.test_code_l4_sub, content_object=self.dummy_order1, user=self.test_client, amount=self.dummy_order1.amount)

        self.test_code_l2_sub2 = TrackCode.objects.create(code="L2SUB2", usage="always", discount_type="perc", discount=20,
                                                         node=self.l2_marketeri2)
        self.cu_side = CodeUse.objects.create(code=self.test_code_l2_sub2, content_object=self.dummy_order1, user=self.test_client, amount=self.dummy_order1.amount)


    def test_min_provision(self):
        assert self.l5_sub.provision_min(100)==22.5, "Min provision not matching expected value"
        assert self.l4_sub.provision_min(100) == 5.625, "Min provision not matching expected value"
        assert self.l3_sub1.provision_min(100) == 1.875, "Min provision not matching expected value"
        assert self.l2_marketeri1.provision_min(100) == 10.0, "Min provision not matching expected value"
        assert self.top_node.provision_min(100) == 40.0, "Min provision not matching expected value"

    def test_max_available(self):
        assert self.l5_sub.max_available(100) == 30.0, "Max provision not matching expected value"
        assert self.l4_sub.max_available(100) == 30.0, "Max provision not matching expected value"
        assert self.l3_sub1.max_available(100) == 30.0, "Max provision not matching expected value"
        assert self.l2_marketeri1.max_available(100) == 40.0, "Max provision not matching expected value"
        assert self.top_node.max_available(100) == 100.0, "Max provision not matching expected value"

    def test_simple_treesplit_no_discount(self):
        cu = self.cu_simple_treesplit_no_discount
        assert cu.amount == cu.control_sum, "Original amount and control sum do not match"
        assert cu.discounted_amount == 0.0, "Discount is not 0"
        assert cu.get_provision() == self.l5_sub.my_cut(self.dummy_order1.amount), "The personal cut should be same as the provision if no discount"
        ap = [x for x in cu.get_ancestor_provisions()]
        assert ap[0][0] == self.top_node      and ap[0][1] == Decimal(600), "Wrong provision or node"
        assert ap[1][0] == self.l2_marketeri1 and ap[1][1] == Decimal(100), "Wrong provision or node"
        assert ap[2][0] == self.l3_sub1       and ap[2][1] == Decimal(18.75), "Wrong provision or node"
        assert ap[3][0] == self.l4_sub        and ap[3][1] == Decimal(56.25), "Wrong provision or node"
        assert ap[4][0] == self.l5_sub        and ap[4][1] == Decimal(225), "Wrong provision or node"

    def test_simple_treesplit_discount(self):
        cu = self.cu_simple_treesplit_discount
        assert cu.amount == cu.control_sum, "Original amount and control sum do not match"
        assert cu.discounted_amount == 100.0, "Discount is not 100"
        ap = [x for x in cu.get_ancestor_provisions()]
        assert ap[0][0] == self.top_node      and ap[0][1] == Decimal(600), "Wrong provision or node"
        assert ap[1][0] == self.l2_marketeri1 and ap[1][1] == Decimal(100), "Wrong provision or node"
        assert ap[2][0] == self.l3_sub1       and ap[2][1] == Decimal(18.75), "Wrong provision or node"
        assert ap[3][0] == self.l4_sub        and ap[3][1] == Decimal(56.25), "Wrong provision or node"
        assert ap[4][0] == self.l5_sub        and ap[4][1] == Decimal(125), "Wrong provision or node"

    def test_simple_treesplit_l4_discount(self):
        cu = self.cu_simple_treesplit_l4_discount
        assert cu.amount == cu.control_sum, "Original amount and control sum do not match"
        assert cu.discounted_amount == 100.0, "Discount is not 100"
        ap = [x for x in cu.get_ancestor_provisions()]
        assert ap[0][0] == self.top_node      and ap[0][1] == Decimal(600), "Wrong provision or node"
        assert ap[1][0] == self.l2_marketeri1 and ap[1][1] == Decimal(100), "Wrong provision or node"
        assert ap[2][0] == self.l3_sub1       and ap[2][1] == Decimal(75), "Wrong provision or node"
        assert ap[3][0] == self.l4_sub        and ap[3][1] == Decimal(125), "Wrong provision or node"

    def sppp(self, node, expected_value):
        node.refresh_from_db()
        val = node.get_sum_provisions_for_period(self.DS, self.DU)
        assert val == expected_value, "%s has sum %.2f, but expected %.2f" % (node, val, expected_value)

    def test_provision_totals(self):
        self.sppp(self.top_node, 2400)
        self.sppp(self.l2_marketeri1, 300)
        self.sppp(self.l3_sub1, 112.5)
        self.sppp(self.l4_sub, 237.5)
        self.sppp(self.l5_sub, 350)


    def test_provision_subtree_l4(self):
        self.l4_sub.refresh_from_db()
        prov_tree = self.l4_sub.get_provisions_tree_for_period(self.DS, self.DU)
        assert prov_tree[0][0] == self.l4_sub and prov_tree[0][2] == 125.0
        assert prov_tree[1][0] == self.l5_sub and prov_tree[1][2] == 112.5

    def test_provision_subtree_l3(self):
        self.l3_sub1.refresh_from_db()
        prov_tree = self.l3_sub1.get_provisions_tree_for_period(self.DS, self.DU)
        assert prov_tree[0][0] == self.l3_sub1 and prov_tree[0][2] == 0 #nothing just for myself
        assert prov_tree[1][0] == self.l4_sub and prov_tree[1][2] == 75
        assert prov_tree[2][0] == self.l5_sub and prov_tree[2][2] == 37.5


class DiscountCodeTestCase(TestCase):
    def setUp(self):
        self.user = User.objects.create_user(username='testuser', password="12345")
        self.user2 = User.objects.create_user(username='testuser2', password="12345")
        self.user3 = User.objects.create_user(username='testuser3', password="12345")
        self.top_node = MarketerNode.objects.create(name="Top Base", split_type="fixed", split_cut=60, owner=self.user )
        self.code_always_1 = TrackCode.objects.create(code="ALWAYS1", usage="always", discount_type="perc", discount=30, node=self.top_node)
        self.code_always_2 = TrackCode.objects.create(code="ALWAYS2", usage="always", discount_type="perc", discount=30, node=self.top_node)
        self.code_useronce_1 = TrackCode.objects.create(code="USER1", usage="useronce", discount_type="perc", discount=30, node=self.top_node)
        self.code_useronce_2 = TrackCode.objects.create(code="USER2", usage="useronce", discount_type="perc", discount=30, node=self.top_node)
        self.code_user_n_2 = TrackCode.objects.create(code="UN2", usage="useronce", discount_type="perc", max_n=2, discount=30, node=self.top_node)
        self.code_once_1 = TrackCode.objects.create(code="ONCE1", usage="once", discount_type="perc", discount=30, node=self.top_node)
        self.code_once_2 = TrackCode.objects.create(code="ONCE2", usage="once", discount_type="perc", discount=30, node=self.top_node)
        self.code_n_2 = TrackCode.objects.create(code="N2", usage="once", discount_type="perc", max_n=2, discount=30, node=self.top_node)


    def test_once(self):
        self.assertEqual((True, 30), CodeUse.test_use(self.code_once_1, self.user, target=None, amount=100))
        CodeUse.use(self.code_once_1, self.user, target=None, amount=100)
        self.assertEqual((False, 30), CodeUse.test_use(self.code_once_1, self.user, target=None, amount=100))
        self.assertEqual((False, 30), CodeUse.test_use(self.code_once_1, self.user2, target=None, amount=100))

        self.assertEqual((True, 30), CodeUse.test_use(self.code_useronce_1, self.user, target=None, amount=100))
        CodeUse.use(self.code_useronce_1, self.user, target=None, amount=100)
        self.assertEqual((False, 30), CodeUse.test_use(self.code_useronce_1, self.user, target=None, amount=100))
        self.assertEqual((True, 30), CodeUse.test_use(self.code_useronce_1, self.user2, target=None, amount=100))

    def test_max_n(self):
        self.assertEqual((True, 30), CodeUse.test_use(self.code_n_2, self.user, target=None, amount=100))
        CodeUse.use(self.code_n_2, self.user, target=None, amount=100)
        self.assertEqual((True, 30), CodeUse.test_use(self.code_n_2, self.user, target=None, amount=100))
        self.assertEqual((True, 30), CodeUse.test_use(self.code_n_2, self.user2, target=None, amount=100))
        CodeUse.use(self.code_n_2, self.user2, target=None, amount=100)
        self.assertEqual((False, 30), CodeUse.test_use(self.code_n_2, self.user, target=None, amount=100))
        self.assertEqual((False, 30), CodeUse.test_use(self.code_n_2, self.user2, target=None, amount=100))

        self.assertEqual((True, 30), CodeUse.test_use(self.code_user_n_2, self.user, target=None, amount=100))
        CodeUse.use(self.code_user_n_2, self.user, target=None, amount=100)
        self.assertEqual((True, 30), CodeUse.test_use(self.code_user_n_2, self.user, target=None, amount=100))
        self.assertEqual((True, 30), CodeUse.test_use(self.code_user_n_2, self.user2, target=None, amount=100))
        CodeUse.use(self.code_user_n_2, self.user2, target=None, amount=100)
        self.assertEqual((True, 30), CodeUse.test_use(self.code_user_n_2, self.user, target=None, amount=100))
        self.assertEqual((True, 30), CodeUse.test_use(self.code_user_n_2, self.user2, target=None, amount=100))
        CodeUse.use(self.code_user_n_2, self.user, target=None, amount=100)
        self.assertEqual((False, 30), CodeUse.test_use(self.code_user_n_2, self.user, target=None, amount=100))
        self.assertEqual((True, 30), CodeUse.test_use(self.code_user_n_2, self.user2, target=None, amount=100))


    def test_always(self):
        self.assertEqual((True, 30), CodeUse.test_use(self.code_always_1, self.user, target=None, amount=100))
        CodeUse.use(self.code_always_1, self.user, target=None, amount=100)
        CodeUse.use(self.code_always_1, self.user, target=None, amount=100)
        CodeUse.use(self.code_always_1, self.user2, target=None, amount=100)
        CodeUse.use(self.code_always_1, self.user2, target=None, amount=100)
        CodeUse.use(self.code_always_1, self.user, target=None, amount=100)
        CodeUse.use(self.code_always_1, self.user, target=None, amount=100)
        CodeUse.use(self.code_always_1, self.user2, target=None, amount=100)
        CodeUse.use(self.code_always_1, self.user2, target=None, amount=100)
        self.assertEqual((True, 30), CodeUse.test_use(self.code_always_1, self.user, target=None, amount=100))


    def test_validity(self):
        self.code_not_yet_valid = TrackCode.objects.create(code="VV1", usage="always", discount_type="perc", discount=30, node=self.top_node,
                                                            valid_from="2100-01-01", valid_until="2199-01-01")
        self.assertEqual((False, 30), CodeUse.test_use(self.code_not_yet_valid, self.user, target=None, amount=100))

        self.code_no_longer_valid = TrackCode.objects.create(code="VV2", usage="always", discount_type="perc", discount=30, node=self.top_node,
                                                            valid_from="1970-01-01", valid_until="1999-01-01")
        self.assertEqual((False, 30), CodeUse.test_use(self.code_no_longer_valid, self.user, target=None, amount=100))
        self.code_valid = TrackCode.objects.create(code="VV3", usage="always", discount_type="perc", discount=30, node=self.top_node,
                                                            valid_from="1970-01-01", valid_until="2999-01-01")
        self.assertEqual((True, 30), CodeUse.test_use(self.code_valid, self.user, target=None, amount=100))



# Generated by Django 3.1.6 on 2021-02-18 18:27

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('mlmtools', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='MarketerNode',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('tn_ancestors_pks', models.TextField(blank=True, default='', editable=False, verbose_name='Ancestors pks')),
                ('tn_ancestors_count', models.PositiveSmallIntegerField(default=0, editable=False, verbose_name='Ancestors count')),
                ('tn_children_pks', models.TextField(blank=True, default='', editable=False, verbose_name='Children pks')),
                ('tn_children_count', models.PositiveSmallIntegerField(default=0, editable=False, verbose_name='Children count')),
                ('tn_depth', models.PositiveSmallIntegerField(default=0, editable=False, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(10)], verbose_name='Depth')),
                ('tn_descendants_pks', models.TextField(blank=True, default='', editable=False, verbose_name='Descendants pks')),
                ('tn_descendants_count', models.PositiveSmallIntegerField(default=0, editable=False, verbose_name='Descendants count')),
                ('tn_index', models.PositiveSmallIntegerField(default=0, editable=False, verbose_name='Index')),
                ('tn_level', models.PositiveSmallIntegerField(default=1, editable=False, validators=[django.core.validators.MinValueValidator(1), django.core.validators.MaxValueValidator(10)], verbose_name='Level')),
                ('tn_priority', models.PositiveSmallIntegerField(default=0, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(9999)], verbose_name='Priority')),
                ('tn_order', models.PositiveSmallIntegerField(default=0, editable=False, verbose_name='Order')),
                ('tn_siblings_pks', models.TextField(blank=True, default='', editable=False, verbose_name='Siblings pks')),
                ('tn_siblings_count', models.PositiveSmallIntegerField(default=0, editable=False, verbose_name='Siblings count')),
                ('name', models.CharField(max_length=256)),
                ('tn_parent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='tn_children', to='mlmtools.marketernode', verbose_name='Parent')),
            ],
            options={
                'ordering': ['tn_order'],
                'abstract': False,
            },
        ),
    ]



# Generated by Django 3.1.6 on 2021-02-20 17:44

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('mlmtools', '0007_auto_20210219_1440'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='trackcode',
            name='owner',
        ),
        migrations.AddField(
            model_name='trackcode',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='trackcode',
            name='discount',
            field=models.DecimalField(decimal_places=2, default=10, max_digits=5),
        ),
        migrations.AddField(
            model_name='trackcode',
            name='discount_type',
            field=models.CharField(choices=[('perc', '%'), ('CZK', 'CZK')], default='perc', max_length=4, verbose_name='dt'),
        ),
        migrations.AddField(
            model_name='trackcode',
            name='node',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='mlmtools.marketernode'),
        ),
        migrations.AddField(
            model_name='trackcode',
            name='usage',
            field=models.CharField(choices=[('always', 'always'), ('useronce', 'once per user'), ('once', 'once only')], default='useronce', max_length=8),
        ),
        migrations.AddField(
            model_name='trackcode',
            name='valid_from',
            field=models.DateField(null=True),
        ),
        migrations.AddField(
            model_name='trackcode',
            name='valid_until',
            field=models.DateField(null=True),
        ),
        migrations.AlterField(
            model_name='marketernode',
            name='split_cut',
            field=models.DecimalField(decimal_places=2, default=100.0, help_text='% of split between THIS one and other', max_digits=5),
        ),
        migrations.AlterField(
            model_name='marketernode',
            name='split_type',
            field=models.CharField(choices=[('fixed', 'fixed'), ('treesplit', 'tree split')], default='fixed', help_text='How to split, fixed amount (1 level), or tree split (recursively sent up the tree)', max_length=16),
        ),
    ]



from datetime import timedelta
from django.contrib.admin.views.decorators import staff_member_required
from django.core.exceptions import PermissionDenied
from django.shortcuts import render
from datetime import datetime
from django.views import View
from mlmtools.models import TrackCode, MarketerNode

def get_granted_node_ids(request):
    ''' Returns list of ids the current user is allowed to see '''
    if request.user.is_superuser:
        return [ x["pk"] for x in MarketerNode.objects.all().values('pk') ]
    node_ids = []
    for node in MarketerNode.objects.filter(owner=request.user):
        node_ids += [node.pk] + [nn.pk for nn in node.descendants]
    return node_ids

@staff_member_required
def list_codes(request):
    node_ids = get_granted_node_ids(request)
    c = {"code_list" : TrackCode.objects.filter(node_id__in=node_ids),
         "granted_node_ids" : node_ids}
    return render(request, "mlmtools/list_codes.html", c)

@staff_member_required
def code_detail(request, code):
    c = { "code" : TrackCode.objects.get(code=code)}
    c["granted_node_ids"] = get_granted_node_ids(request)
    if c["code"].node.pk not in c["granted_node_ids"]:
        raise PermissionDenied("not allowed to view this node")
    if request.GET.get("since"):
        request.session["since"] = request.GET.get("since")
    if request.GET.get("until"):
        request.session["until"] = request.GET.get("until")
    c["since"] = request.session.get("since", (datetime.now().date() - timedelta(days=30)).isoformat())
    c["until"] = request.session.get("until", (datetime.now().date() + timedelta(days=3)).isoformat())
    c["usages"] = c["code"].codeuse_set.filter(used_at__gte=c["since"], used_at__lt=c["until"]).order_by("-used_at")
    return render(request, "mlmtools/code_detail.html", c)

@staff_member_required
def node_detail(request, pk):
    c = {"node" : MarketerNode.objects.get(id=pk)}
    c["granted_node_ids"] = get_granted_node_ids(request)
    if c["node"].pk not in c["granted_node_ids"]:
        raise PermissionDenied("not allowed to view this node")
    if request.GET.get("since"):
        request.session["since"] = request.GET.get("since")
    if request.GET.get("until"):
        request.session["until"] = request.GET.get("until")
    c["since"] = request.session.get("since", (datetime.now().date() - timedelta(days=30)).isoformat())
    c["until"] = request.session.get("until", (datetime.now().date() + timedelta(days=3)).isoformat())
    c["provisions"] = c["node"].get_provisions_for_period(c["since"], c["until"])
    c["provisions_tree"] = c["node"].get_provisions_tree_for_period(c["since"], c["until"])
    c["total_provisions"] = c["node"].get_sum_provisions_for_period(c["since"], c["until"])
    return render(request, "mlmtools/node_detail.html", c)



# -*- coding: utf-8 -*-
import floppyforms
from django import forms
from treenode.utils import split_pks
from .models import MarketerNode




class MarketerNodeForm(forms.ModelForm):
    split_cut = forms.DecimalField(min_value=0, max_value=100)
    class Meta:
        model = MarketerNode
        exclude = []

        # def __init__(self, *args, **kwargs):
        #     super(MarketerNodeForm, self).__init__(*args, **kwargs)
        # if 'tn_parent' not in self.fields:
        #     return
        # exclude_pks = []
        # obj = self.instance
        # if obj.pk:
        #     exclude_pks += [obj.pk]
        #     exclude_pks += split_pks(obj.tn_descendants_pks)
        # manager = obj.__class__.objects
        # self.fields['tn_parent'].queryset = manager.exclude(pk__in=exclude_pks)



# Generated by Django 3.2.2 on 2021-05-13 09:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('mlmtools', '0009_auto_20210305_1743'),
    ]

    operations = [
        migrations.CreateModel(
            name='DummyTarget',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=2048)),
                ('amount', models.DecimalField(decimal_places=2, max_digits=5)),
            ],
        ),
    ]



# Generated by Django 3.1.6 on 2021-02-19 13:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('mlmtools', '0003_auto_20210219_0836'),
    ]

    operations = [
        migrations.AddField(
            model_name='marketernode',
            name='can_have_children',
            field=models.BooleanField(default=True),
        ),
        migrations.AddField(
            model_name='marketernode',
            name='parent_cut',
            field=models.DecimalField(decimal_places=2, default=50.0, help_text='How much to sent up the tree', max_digits=5),
        ),
        migrations.AddField(
            model_name='marketernode',
            name='split_type',
            field=models.CharField(choices=[('fixed', 'fixed'), ('tree split', 'treesplit')], default='fixed', help_text='How to split, fixed amount (1 level), or tree split (recursively sent up the tree)', max_length=16),
        ),
    ]



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mlmprovisions", # Replace with your own username
    version="0.1.8",
    author="Belda",
    author_email="jakub.belescak@centrum.cz",
    description="Reusable Django app to track provisions for affiliate partners. Includes tools for multi-level marketing provisions.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/belda/mlmprovisions",
    packages=['mlmtools'],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.4',
    install_requires = [
        "django",
        "django-floppyforms",
        "django-treenode"
    ]
)


# Generated by Django 3.2.2 on 2021-11-12 10:16

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('mlmtools', '0011_alter_dummytarget_amount'),
    ]

    operations = [
        migrations.AlterField(
            model_name='codeuse',
            name='amount',
            field=models.DecimalField(decimal_places=2, default=10, max_digits=10),
        ),
    ]



# Generated by Django 3.2.3 on 2022-09-02 12:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('mlmtools', '0012_alter_codeuse_amount'),
    ]

    operations = [
        migrations.AddField(
            model_name='trackcode',
            name='max_n',
            field=models.PositiveIntegerField(default=1, help_text='The maximum number of codes applied to usage choices'),
        ),
        migrations.AlterField(
            model_name='trackcode',
            name='usage',
            field=models.CharField(choices=[('always', 'always'), ('useronce', 'Nx / user'), ('once', 'Nx / user')], default='useronce', max_length=8),
        ),
    ]



from django.conf.urls import url, include
from mlmtools.views import list_codes, code_detail, node_detail

urlpatterns = [
    url("^list_codes/", list_codes, name="list_codes"),
    url("^code_detail/(?P<code>[A-Za-z0-9]{4,15})/", code_detail, name="code_detail"),
    url("^node_detail/(?P<pk>[0-9]+)/", node_detail, name="node_detail"),
]


# Generated by Django 3.1.6 on 2021-03-05 17:43

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0002_remove_content_type_name'),
        ('mlmtools', '0008_auto_20210220_1744'),
    ]

    operations = [
        migrations.AlterField(
            model_name='trackcode',
            name='usage',
            field=models.CharField(choices=[('always', 'always'), ('useronce', '1x / user'), ('once', 'once only')], default='useronce', max_length=8),
        ),
        migrations.AlterField(
            model_name='trackcode',
            name='valid_from',
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='trackcode',
            name='valid_until',
            field=models.DateField(blank=True, null=True),
        ),
        migrations.CreateModel(
            name='CodeUse',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.PositiveIntegerField(null=True)),
                ('used_at', models.DateTimeField(auto_now=True, db_index=True)),
                ('amount', models.DecimalField(decimal_places=2, default=10, max_digits=5)),
                ('code', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, to='mlmtools.trackcode')),
                ('content_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='contenttypes.contenttype')),
                ('user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.DO_NOTHING, to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]
